#include <iostream>
#include <vector>
#include <map>
#include <algorithm>
#include "Matrix.h"
#include "RBM.h"
#include "Random.h"
#include "QRDecomposition.h"
#include "NaturalGradient.h"

using namespace clap;

//_____________________________________________________________________________
//                                                                            |
//                                CONSTRUCTOR                                 |
// ___________________________________________________________________________|

NaturalGradient::NaturalGradient (int vd, int hd, int ms, int nfs, int ngs, double lr, bool un) : rbm (vd, hd) {
  USE_NATURAL_GRADIENT = un;
  vDim = vd;
  hDim = hd;
  markovSamples = ms;
  numFisherSamples = nfs;
  numGradientSamples = ngs;
  learningRate = lr;
  stopped = false;
  fisherValidationError = false;
  fisherInversionError = false;
}

std::pair< clap::Matrix<double>, clap::Matrix<double> > NaturalGradient::sampleFromRBM(const clap::RBM<double>& samplingRBM) {
    clap::Matrix<double> x = clap::Matrix<double>::createNew (vDim, 1);
    // 1. 1. start with a random binary vector
    for (int i = 0; i < vDim; ++i) {
      x (i) = rnd.bernoulli (0.5);
    }
    // 1. 2. do Gibbs sampling
    clap::Matrix<double> h = clap::Matrix<double>::createNew (hDim, 1);
    for (int k = 0; k < markovSamples; ++k) {
      h = samplingRBM.sampleHiddenGivenVisible (x);
      x = samplingRBM.sampleVisibleGivenHidden (h);
    }
    return std::make_pair(x,h);
}

void NaturalGradient::fillSamples (const clap::RBM<double>& samplingRBM, const clap::RBM<double>& rbm, int numSamples, std::vector<Sample>* samples) {
  samples->clear();
  // 1. generate samples
  for (int s = 0; s < numSamples; ++s) {
    std::pair< clap::Matrix<double>, clap::Matrix<double> > p = sampleFromRBM(samplingRBM);
    // 1. 3. keep x and h for computing J and set f to 0
    samples->push_back ( Sample ( p.first, p.second, 0) );
  }

}

//_____________________________________________________________________________
//                                                                            |
//                 COMPUTE THE INVERSE FISHER INFORMATION MATRIX              |
// ___________________________________________________________________________|

Matrix<double> NaturalGradient::computeFisher (const std::vector<Sample>& samples, clap::Matrix<double>& Ex2, clap::Matrix<double>& Ex) {
  // compute Fisher Information Matrix
  int thetaSize = hDim * vDim + vDim + hDim;
  Ex2 = clap::Matrix<double>::createZero ( thetaSize, thetaSize );
  Ex = clap::Matrix<double>::createZero ( thetaSize, 1 );
  for (unsigned int i = 0; i < samples.size(); ++i) {
    const Sample& s = samples[i];
    clap::Matrix<double> weightsUpdate = s.h * s.v.transpose();
    clap::Matrix<double> thetaUpdate = weightsUpdate.resizedProxy (vDim * hDim, 1) || s.v || s.h;
    Ex2 += (thetaUpdate * thetaUpdate.transpose() );
    Ex += thetaUpdate;
  }
  clap::Matrix<double> Ex2f = Ex2 * (1. / samples.size());
  clap::Matrix<double> Exf  = Ex  * (1. / samples.size());
  return Ex2f - Exf * Exf.transpose();
}


