#ifndef OPTIM_RBMS_SRC_NATURAL_GRADIENT_H
#define OPTIM_RBMS_SRC_NATURAL_GRADIENT_H

#include <iostream>
#include <vector>
#include <map>
#include <algorithm>
#include "Matrix.h"
#include "RBM.h"
#include "Random.h"
#include "QRDecomposition.h"
#include "EigenvalueDecomposition.h"

struct NaturalGradient {

//_____________________________________________________________________________
//                                                                            |
//                        ALGORITHM PARAMETERS & INPUTS                       |
// ___________________________________________________________________________|


  /// Size of the input dimension = visible dimension.
  int vDim;

  /// Size of the hidden dimension.
  int hDim;

  /// Number of gibbs iterations to use for
  /// sampling points from the RBM.
  int markovSamples;

  /// Number of function samples to use at each iteration. This should be
  /// greater that ... for the fisher information matrix to be invertible.
  int numFisherSamples;

  /// Number of function samples to use at each iteration. This should be
  /// greater that ... for the fisher information matrix to be invertible.
  int numGradientSamples;

  /// Learning rate;
  double learningRate;

  /// Random number generator.
  clap::Random<double> rnd;

  /// True if the optimization procedure stopped
  bool stopped;

  /// True if optimization stopped because of a cross-validation error
  bool fisherValidationError;

  /// True if optimization stopped because of Fisher matrix singularity
  bool fisherInversionError;

  inline bool hasStopped()const{ return stopped; }

  bool USE_NATURAL_GRADIENT;

  /// Create a NaturalGradient algorithm and set the parameters.
  /// @arg vd dimension of the input vector.
  /// @arg hd hidden dimension of the sampling RBM.
  /// @arg ms number of iterations of gibbs sampling to use when sampling from
  /// the RBM.
  /// @arg nfs number of samples to take for the estimation of the fisher
  /// information matrix.
  /// @arg ngs number of samples to take for the estimation of the gradient.
  NaturalGradient (int vd, int hd, int ms, int nfs, int ngs, double lr, bool un);

//_____________________________________________________________________________
//                                                                            |
//                                SAMPLE STRUCT                               |
// ___________________________________________________________________________|

  /// All the information about a sample.
  struct Sample {
    clap::Matrix<double> v;
    clap::Matrix<double> h;
    double fvalue;
    Sample (clap::Matrix<double> cv, clap::Matrix<double> ch, double cfvalue) :
      v (cv), h (ch), fvalue (cfvalue) {}
    bool operator< (const Sample& s) const {
      return fvalue > s.fvalue;
    }
  };

//_____________________________________________________________________________
//                                                                            |
//                         GENERATE SAMPLES FROM RBM                          |
// ___________________________________________________________________________|

  /// Generate one sample from the given RBM.
  std::pair< clap::Matrix<double>, clap::Matrix<double> > sampleFromRBM(const clap::RBM<double>& samplingRBM);
  /// Fill the samples vector with SAMPLE_SIZE Samples generated from the given
  /// RBM using temperature 1/beta. The generated samples are returned in the
  /// samples vector.
  /// @arg samplingRBM the regularized rbm from which to draw samples.
  /// @arg rbm the target RBM for wich we should be drawing.
  /// @arg beta the inverse temperature of the RBM.
  /// @arg samples a vector of Sample to  fill.
  void fillSamples (const clap::RBM<double>& samplingRBM, const clap::RBM<double>& rbm, int numSamples, std::vector<Sample>* samples);

  /// @arg f the target function of the optimization procedure.
  /// @arg samples a vector of filled Samples to evaluate with the target function f.
  template<class F>
  void evaluateSamples (F f, std::vector<Sample>* samples);

//_____________________________________________________________________________
//                                                                            |
//                                WEIGHT SAMPLES                              |
// ___________________________________________________________________________|


  /// Sort the given samples and give them a weight according to the function q
  template<class Q>
  std::vector<double> weightSamples (Q q, const std::vector<Sample>& samples) ;

//_____________________________________________________________________________
//                                                                            |
//                 COMPUTE THE INVERSE FISHER INFORMATION MATRIX              |
// ___________________________________________________________________________|

  /// @return the Fisher information matrix estimated over the given sample.
  clap::Matrix<double> computeFisher (const std::vector<Sample>& samples, clap::Matrix<double>& Ex2, clap::Matrix<double>& Ex);
  clap::Matrix<double> computeFisher2 (const std::vector<Sample>& samples);

  /// Optimize the function f with the weight function q using an RBM and the
  /// natural gradient.
  template<class F, class Q>
  void optimizeStep (F f, Q q);

//_____________________________________________________________________________
//                                                                            |
//                         ALGORITHM INTERMEDIARY DATA                        |
// ___________________________________________________________________________|

  /// rbm used to sample points
  clap::RBM<double> rbm;

  /// samples (gradient samples when out of optimize func).
  std::vector<Sample> gsamples;

}; // class NaturalGradient

//_____________________________________________________________________________
//                                                                            |
//                        INLINE & TEMPLATE DEFINITIONS                       |
// ___________________________________________________________________________|

template<class F>
void NaturalGradient::evaluateSamples (F f, std::vector<Sample>* samples) {
  // 1. Evaluate samples with f
  for (int i = 0; i < samples->size(); ++i) {
    Sample& s = (*samples) [i];
    s.fvalue = f (s.v);
  }
  // 2. sort samples in increasing order
  std::sort (samples->begin(), samples->end() );
}

template<class Q>
std::vector<double> NaturalGradient::weightSamples (Q q, const std::vector<Sample>& samples) {

  std::vector<double> sampleWeights;

  for (int s = 0; s < samples.size(); ++s) {
    int ties = 0;
    while ( s + ties + 1 < samples.size() && samples[s+ties+1].fvalue == samples[s].fvalue ) {
      ties += 1;
    }
    double expectationOverTies = 0;
    for (int i = 0; i < ties + 1; ++i) {
      expectationOverTies += q (s + i);
    }
    expectationOverTies /= ties + 1;
    for (int i = 0; i < ties + 1; ++i) {
      sampleWeights.push_back (expectationOverTies);
    }
    s += ties;
  }

  return sampleWeights;
}

template<class F, class Q>
void NaturalGradient::optimizeStep (F f, Q q) {

  clap::Matrix<double> Finv;
  int thetaSize = hDim * vDim + vDim + hDim;

  // COMPUTE FISHER INFORMATION MATRIX

  bool fisherSuccess = false;
  if ( USE_NATURAL_GRADIENT ) {
    try {

      // compute the fisher matrix
      std::vector<Sample> fsamples1;
      clap::Matrix<double> E1x2, E1x;
      fillSamples (rbm, rbm, numFisherSamples/2, &fsamples1);
      clap::Matrix<double> Fish1 = computeFisher (fsamples1,E1x2,E1x);
      // compute the fisher matrix a second time
      std::vector<Sample> fsamples2;
      clap::Matrix<double> E2x2, E2x;
      fillSamples (rbm, rbm, numFisherSamples - numFisherSamples/2, &fsamples2);
      clap::Matrix<double> Fish2 = computeFisher (fsamples2,E2x2,E2x);

      // compute first inverse
      clap::QRDecomposition<double> qr1 (clap::Matrix<double>::createCopy (Fish1) );
      clap::Matrix<double> Finv1 = clap::Matrix<double>::createCopy (qr1.solve (clap::Matrix<double>::createIdentity (thetaSize, thetaSize ) ) );
      // compute second inverse
      clap::QRDecomposition<double> qr2 (clap::Matrix<double>::createCopy (Fish2) );
      clap::Matrix<double> Finv2 = clap::Matrix<double>::createCopy (qr2.solve (clap::Matrix<double>::createIdentity (thetaSize, thetaSize ) ) );

      // compute F2F1m1 and F1F2m1
      clap::Matrix<double> F2F1m1 = clap::Matrix<double>::createCopy (Fish2) *clap::Matrix<double>::createCopy (Finv1);
      clap::Matrix<double> F1F2m1 = clap::Matrix<double>::createCopy (Fish1) *clap::Matrix<double>::createCopy (Finv2);

      // (F_1^{-1/2}F_2F_1^{-1/2} - Id) qui est égale à
      // Tr( ( F_2 F_1^{-1} )^2) -2 Tr( F_2 F_1^{-1} )+Tr( Id )

      double Tr1 = 0.;
      double Tr2 = 0.;
      for (int i = 0; i < thetaSize; ++i){
        for (int j = 0; j < thetaSize; ++j){
          Tr1 += F2F1m1(j,i) * F2F1m1(i,j);
          Tr2 += F1F2m1(j,i) * F1F2m1(i,j);
        }
      }

      double freezeCriterion1 = Tr1 - 2 * F2F1m1.trace() + thetaSize;
      double freezeCriterion2 = Tr2 - 2 * F1F2m1.trace() + thetaSize;

      // Test equivalence with base criterion: passed
      //ASSERT ( abs( freezeCriterion1 - ((F2F1m1 * F2F1m1).trace() - 2 * F2F1m1.trace() + F2F1m1.rows()) ) < 1e-5 );
      //ASSERT ( abs( freezeCriterion2 - ((F1F2m1 * F1F2m1).trace() - 2 * F1F2m1.trace() + F1F2m1.rows()) ) < 1e-5 );
      ASSERT ( freezeCriterion1 >= 0 && freezeCriterion2 >= 0);

      if ( freezeCriterion1 < 1.0*thetaSize && freezeCriterion2 < 1.0*thetaSize ) {

        fisherSuccess = true;

        // recompose Fisher matrix
        clap::Matrix<double> Ex2 = E1x2 + E2x2;
        clap::Matrix<double> Ex = E1x + E2x;
        Ex2 *= 1./numFisherSamples;
        Ex  *= 1./numFisherSamples;
        clap::Matrix<double> Fish = Ex2 - Ex*Ex.transpose();

        // Test euquivalence of Fisher matrix recomposision: passed
        //fsamples1.insert(fsamples1.end(), fsamples2.begin(), fsamples2.end());
        //ASSERT(fsamples1.size() == numFisherSamples);
        //clap::Matrix<double> Ex2Test, ExTest;
        //clap::Matrix<double> FishTest = computeFisher (fsamples1, Ex2Test, ExTest);
        //std::cout << (Fish - FishTest).l1Norm() << std::endl;

        // inverse Fisher matrix
        clap::QRDecomposition<double> qr (clap::Matrix<double>::createCopy (Fish) );
        Finv = clap::Matrix<double>::createCopy (qr.solve (clap::Matrix<double>::createIdentity (thetaSize, thetaSize ) ) );

      } else {
        fisherValidationError = true;
        std::cout << "Matrices are not near enough" << std::endl;
      }
    } catch (plugscape::RuntimeError ex) {
      fisherInversionError = true;
      std::cout << "Inversion failed" << ex.what() << std::endl;
    }
    if (!fisherSuccess) {
      stopped = true;
      std::cout << "trying again...NOT" << std::endl;
      return;
    }
  }

  // fill the gradient samples using the current rbm
  fillSamples (rbm, rbm, numGradientSamples, &gsamples);

  // WEIGHT SAMPLES ACCORDING TO QUANTILES

  evaluateSamples (f,&gsamples);

  clap::Matrix<double> posWeightsUpdate = clap::Matrix<double>::createZero (hDim, vDim);
  clap::Matrix<double> posVbiasUpdate = clap::Matrix<double>::createZero (vDim, 1);
  clap::Matrix<double> posHbiasUpdate = clap::Matrix<double>::createZero (hDim, 1);
  clap::Matrix<double> negWeightsUpdate = clap::Matrix<double>::createZero (hDim, vDim);
  clap::Matrix<double> negVbiasUpdate = clap::Matrix<double>::createZero (vDim, 1);
  clap::Matrix<double> negHbiasUpdate = clap::Matrix<double>::createZero (hDim, 1);

  std::vector<double> W = weightSamples (q, gsamples);

  // COMPUTE GRADIENT

  // Compute < W(xs) GRAD[ log P(xs) ] >
  for (int s = 0; s < numGradientSamples; ++s) {
    //std::cout << "(" << msamples[s].v.l1Norm() << "," << msamples[s].samplingWeight << "); "<< std::flush;
    // positive term
    posWeightsUpdate += (gsamples[s].h * gsamples[s].v.transpose() ) * W[s];
    posVbiasUpdate   += (gsamples[s].v) * W[s];
    posHbiasUpdate   += (gsamples[s].h) * W[s];
    // negative term
    negWeightsUpdate += (gsamples[s].h * gsamples[s].v.transpose() );
    negVbiasUpdate   += (gsamples[s].v);
    negHbiasUpdate   += (gsamples[s].h);
  }

  //std::cout << std::endl;

  double sumW = 0;
  for (int i = 0; i < numGradientSamples; ++i) sumW += W[i];
  double isumW = 1. / sumW;

  clap::Matrix<double> WeightsUpdate = clap::Matrix<double>::createZero (hDim, vDim);
  clap::Matrix<double> VbiasUpdate = clap::Matrix<double>::createZero (vDim, 1);
  clap::Matrix<double> HbiasUpdate = clap::Matrix<double>::createZero (hDim, 1);
  WeightsUpdate = posWeightsUpdate * (isumW) - negWeightsUpdate * (1./numGradientSamples) ;
  VbiasUpdate = posVbiasUpdate * (isumW) - negVbiasUpdate * (1./numGradientSamples);
  HbiasUpdate = posHbiasUpdate * (isumW) - negHbiasUpdate * (1./numGradientSamples);

  clap::Matrix<double> espgradAsVector = clap::Matrix<double>::createNew (vDim * hDim + vDim + hDim, 1);
  for (int i = 0; i < WeightsUpdate.size(); ++i) espgradAsVector (i) = WeightsUpdate (i);
  for (int i = 0; i < vDim; ++i) espgradAsVector (vDim * hDim + i) = VbiasUpdate (i);
  for (int i = 0; i < hDim; ++i) espgradAsVector (vDim * hDim + vDim + i) = HbiasUpdate (i);

  clap::Matrix<double> updateAsVector;
  if(USE_NATURAL_GRADIENT) updateAsVector = Finv * espgradAsVector;
  else updateAsVector = espgradAsVector;

  for (int i = 0; i < WeightsUpdate.size(); ++i) WeightsUpdate (i) = updateAsVector (i);
  for (int i = 0; i < vDim; ++i) VbiasUpdate (i) = updateAsVector (vDim * hDim + i);
  for (int i = 0; i < hDim; ++i) HbiasUpdate (i) = updateAsVector (vDim * hDim + vDim + i);

  rbm.weights += WeightsUpdate * learningRate;
  rbm.vbias += VbiasUpdate * learningRate;
  rbm.hbias += HbiasUpdate * learningRate;

}

#endif // OPTIM_RBMS_SRC_NATURAL_GRADIENT_H

