#ifndef OPTIM_RBMS_SRC_NATURAL_GRADIENT_EXPERIMENTS_H
#define OPTIM_RBMS_SRC_NATURAL_GRADIENT_EXPERIMENTS_H

#include <omp.h>
#include <sstream>
#include "NaturalGradient.h"
#include "Matrix.h"
#include "Experiment.h"
#include "Statistics.h"

using namespace clap;

//________________________________________________________
//                                                        |
//         COST FUNCTIONS TO MINIMIZE                     |
// _______________________________________________________|

/// Optimum at 0 + Optimum at 111...1.
struct BisphereFunctor {

  Matrix<double>& goptimum;

  BisphereFunctor(Matrix<double>& opt) : goptimum(opt) {}

  double operator() (clap::Matrix<double> x){
    ASSERT( goptimum.size() == x.size() );
    clap::Matrix<double> diff = goptimum - x;
    return std::min ( diff.l1Norm(), diff.size() - diff.l1Norm() );
  }

};

/// Weight function for samples rank ( ex: rank(best point) = 0). This
/// function needs to be defined for all integers between 0 and
/// numSamples - 1 (inclusive).
/// does numSamples need to be even for this to work nicely ?
struct BinaryRankingFunctor {

  int mu;

  BinaryRankingFunctor (int m) {
    mu = m;
  }

  double operator() (int rank) {
    if (rank < mu) return 0;
    else return 1;
  }

};

//__________________________________________________________
//                                                          |
//              BISPHERE DIVERSITY                          |
// _________________________________________________________|

struct bisphere_diversity : public clap::Experiment {

  bisphere_diversity() : clap::Experiment ("Bisphere Diversity-1.3") {
    setParameter ("nbRuns", 100);
    setParameter ("nbSteps", 100);
    setParameter ("vDim", 40);
    setParameter ("hDim", 1);
    setParameter ("markovSamples", 50);
    setParameter ("numFisherSamples", 10000);
    setParameter ("numGradientSamples", 10000);
    setParameter ("learningRate", 0.25);
    setParameter ("natural", true);
    setParameter ("keepSingleRuns", true);
    setParameter ("numSelection", 2000);
    setParameter ("initialNoiseDev", .01 / getIntParameter ("vDim") );
    ASSERT ( getIntParameter ("numSelection") < getIntParameter ("numGradientSamples") );
  }

  void run() {

    std::vector<clap::Statistics<double> > stats (getIntParameter ("nbSteps") );
    std::vector<clap::Statistics<double> > hstats (getIntParameter ("nbSteps") );
    std::vector<clap::Statistics<double> > mstats (getIntParameter ("nbSteps") );

    int fisherValidationErrors = 0;
    int fisherInversionErrors = 0;

    int run;
    # pragma omp parallel for schedule (dynamic)  shared(fisherValidationErrors,fisherInversionErrors,stats,hstats,mstats) private(run)
    for (run = 0; run < getIntParameter ("nbRuns"); ++run) {

      //getfile ("log") << "run(" << run << ")" << std::endl;

      try {

        //getfile ("log") << "set openmp env" << std::endl;

        int tid = omp_get_thread_num();
        std::stringstream sstid;
        sstid << tid;
        std::string stid = std::string (sstid.str() );

        //#pragma omp critical(log1)
        //{
        //  getfile ("log") << "T(" << tid << ")" << "Starting " << (getBoolParameter ("natural") ? "natural" : "vanilla")
        //  << " run no " << run << std::endl;
        //}

        getfile ("log") << "T(" << tid << ")" << "Setting up Natural Gradient" << std::endl;

        NaturalGradient ng (
          getIntParameter ("vDim"),
          getIntParameter ("hDim"),
          getIntParameter ("markovSamples"),
          getIntParameter ("numFisherSamples"),
          getIntParameter ("numGradientSamples"),
          getDoubleParameter ("learningRate"),
          getBoolParameter ("natural")
        );

        // initialize RBM
        // weights
        for (int i = 0; i < getIntParameter ("vDim"); ++i) {
          for (int j = 0; j < getIntParameter ("hDim"); ++j) {
            ng.rbm.weights (j, i) = ng.rbm.rnd.normal (0, 1. / sqrt (getIntParameter ("vDim") * getIntParameter ("hDim") ) );
          }
        }
        // vbias
        for (int i = 0; i < getIntParameter ("vDim"); ++i) {
          double sumj = 0;
          for (int j = 0; j < getIntParameter ("hDim"); ++j) sumj += ng.rbm.weights (j, i);
          ng.rbm.vbias (i) = - (sumj * 0.5);
        }
        // hbias
        for (int j = 0; j < getIntParameter ("hDim"); ++j) {
          double sumi = 0;
          for (int i = 0; i < getIntParameter ("vDim"); ++i) sumi += ng.rbm.weights (j, i);
          ng.rbm.hbias (j) = - (sumi * 0.5);
        }
        // little noise on vbias
        for (int k = 0; k < getIntParameter ("vDim"); ++k) {
          ng.rbm.vbias (k) += ng.rnd.normal (0., getDoubleParameter ("initialNoiseDev") );
        }

        // initialize optimum
        Matrix<double> goptimum = clap::Matrix<double>::createNew ( getIntParameter ("vDim"), 1 );
        for (int k = 0; k < getIntParameter ("vDim"); ++k) {
          goptimum (k) = ng.rnd.bernoulli (0.5);
        }

        // initialize functors
        BinaryRankingFunctor binaryRanking (getIntParameter ("numSelection") );
        BisphereFunctor bisphere (goptimum);

        // declare variables for keeping stats
        std::vector<std::vector<double> > l1points (getIntParameter ("nbSteps"), std::vector<double>() );
        std::vector<NaturalGradient::Sample> testSamples;

        for (int j = 0; j < getIntParameter ("nbSteps"); ++j) {

          // if optimization has not stopped:
          //   - optimize one step
          //   - update test samples
          //   - check if there was an error just now
          bool juststopped = false;
          if ( ! ng.hasStopped() ) {
            ng.optimizeStep (bisphere, binaryRanking);
            testSamples.clear();
            testSamples = ng.gsamples;

            ASSERT( ng.stopped == (ng.fisherInversionError || ng.fisherValidationError) );
            if (ng.stopped) juststopped = true;
          }

          ASSERT ( testSamples.size() == getIntParameter ("numGradientSamples") );
          std::vector<clap::Matrix<double> > diffs;
          for (int n = 0; n < getIntParameter ("numGradientSamples"); ++n) {
            diffs.push_back ( testSamples[n].v - goptimum );
          }

          // split in points closer to the first optimum of the second
          std::vector<NaturalGradient::Sample> group1, group2;
          for (int n = 0; n < getIntParameter ("numGradientSamples"); ++n) {
            if ( diffs[n].l1Norm() <= getIntParameter ("vDim") / 2) {
              ASSERT ( diffs[n].l1Norm() == testSamples[n].fvalue );
              group1.push_back (testSamples[n]);
            } else {
              ASSERT ( getIntParameter ("vDim") - diffs[n].l1Norm() == testSamples[n].fvalue );
              group2.push_back (testSamples[n]);
            }
          }
          std::sort (group1.begin(), group1.end() );
          std::sort (group2.begin(), group2.end() );

          int n1 = group1.size();
          double worst1 = -1;
          double med1 = -1;
          double best1 = -1;
          if ( group1.size() > 0 ) {
            worst1 = group1[0].fvalue;
            med1 = group1[group1.size() /2].fvalue;
            best1 = group1[group1.size()-1].fvalue;
          }
          int n2 = group2.size();
          double worst2 = -1;
          double med2 = -1;
          double best2 = -1;
          if ( group2.size() > 0 ) {
            worst2 = group2[0].fvalue;
            med2 = group2[group2.size() /2].fvalue;
            best2 = group2[group2.size()-1].fvalue;
          }
          if ( best1 == -1 ) best1 = getIntParameter ("vDim") - worst2;
          if ( best2 == -1 ) best2 = getIntParameter ("vDim") - worst1;

          #pragma omp critical(log2)
          {
            getfile ("log") << "T(" << tid << ")" << j << "\t" << n1 << "\t" << worst1 << "\t" << med1 << "\t" << best1 << "\t" << n2 << "\t" << worst2 << "\t" << med2 << "\t" << best2 << std::endl;
            getfile ("sr1-" + stid) << ( (best1 < best2) ? best1 : best2) << " " << std::flush;
            getfile ("sr2-" + stid) << ( (best1 < best2) ? best2 : best1) << " " << std::flush;
          }

          // compute statistics
          double nhone = 0;
          for (int n = 0; n < getIntParameter ("numGradientSamples"); ++n) {
            double l1norm = diffs[n].l1Norm();
            l1points[j].push_back (l1norm);
            nhone += testSamples[n].h (0);
          }
          #pragma omp critical(hstats1)
          {
            hstats[j].add (nhone / getIntParameter ("numGradientSamples") );
          }
          std::sort (l1points[j].begin(), l1points[j].end() );
          //std::sort(fpoints[j].begin(),fpoints[j].end());
          int lowest  = * (l1points[j].begin() );
          int highest = * (l1points[j].end() - 1);
          int highdist = getIntParameter ("vDim") - highest;
          #pragma omp critical(stats1)
          {
            if ( highdist < lowest ) {
              stats[j].add ( highdist );
              mstats[j].add ( lowest );
            } else {
              stats[j].add ( lowest );
              mstats[j].add ( highdist );
            }
          }
          std::vector<double>::iterator it = std::unique (l1points[j].begin(), l1points[j].end() );
          l1points[j].resize (it - l1points[j].begin() );

          #pragma omp critical(errors)
          { // if a run is frozen without having reached one of the two optima
            if (juststopped && (lowest != 0 && highdist != 0) ){
              if (ng.fisherValidationError) fisherValidationErrors += 1;
              if (ng.fisherInversionError) fisherInversionErrors += 1;
            }
          }

        }

        #pragma omp critical(sr1)
        {
          getfile ("sr1-" + stid) << std::endl;
          getfile ("sr2-" + stid) << std::endl;
        }

      } catch (std::bad_alloc e) {
        std::cerr << "bad_alloc was thrown inside parallel block.\nExiting program.";
        exit (0);
      } catch (...) {
        std::cerr << "Unhandled Exception was thrown inside parallel block.\nExiting program.";
        exit (0);
      }
    }

    getfile ("stats") << "step avg var" << std::endl;

    for (int j = 0; j < getIntParameter ("nbSteps"); ++j) {
      if (stats[j].size() != 0) getfile ("stats") << j << " " << stats[j].median() << " " << stats[j].quantile (0.16) << " " << stats[j].quantile (0.84) << std::endl;
    }

    getfile ("mstats") << "step mavg var" << std::endl;

    for (int j = 0; j < getIntParameter ("nbSteps"); ++j) {
      if (mstats[j].size() != 0) getfile ("mstats") << j << " " << mstats[j].median() << " " << mstats[j].quantile (0.16) << " " << mstats[j].quantile (0.84) << std::endl;
    }

    getfile ("hstats") << "step havg hvar" << std::endl;

    for (int j = 0; j < getIntParameter ("nbSteps"); ++j) {
      if (hstats[j].size() != 0) getfile ("hstats") << j << " " << hstats[j].median() << " " << hstats[j].quantile (0.16) << " " << hstats[j].quantile (0.84) << std::endl;
    }

    getfile ("log") << "fisherValidationErrors" << fisherValidationErrors << std::endl;
    getfile ("log") << "fisherInversionErrors" << fisherInversionErrors << std::endl;

  }

};


#endif // OPTIM_RBMS_SRC_NATURAL_GRADIENT_EXPERIMENTS_H

