#include "NaturalGradient.h"

#include <algorithm>
#include "gtest/gtest.h"
#include "Matrix.h"
#include "experiments.h"
#include "plugscape/benchmark.h"

//_____________________________________________________________________________
//                                                                            |
//                                 AUTOMATIC TESTS                            |
// ___________________________________________________________________________|

// Another way to compute the Fisher matrix to check that we find the same thing.
clap::Matrix<double> NaturalGradient::computeFisher2 (const std::vector<Sample>& samples) {
// Weights of the RBM are now refered to as THETA and consist of
  // a single vector wich can be decomposed as follows
  //
  //  THETA = [Wc1] [Wc2] [Wc3] ...   [VB]   [HB]
  //
  // The Fisher Information Matrix can be decomposed into blocks as follows
  //
  //            W           VB         HB
  //       c1 c2 c3 ... |_________|__________|
  //    c1              |         |          |
  //    c2              |         |          |
  //  W c3     FWW      |  FWVB   |   FWHB   |
  //    ...             |         |          |
  //    ..._____________|_________|__________|
  //      |             |         |          |
  //  VB  |    FWVB'    |  FVBVB  |  FVBHB   |
  //     _|_____________|_________|__________|
  //      |             |         |          |
  //  HB  |    FWHB'    |  FVBHB' |  FHBHB   |
  //     _|_____________|_________|__________|

  // FWW = COV ( Wab, Wcd )
  clap::Matrix<double> FWW = clap::Matrix<double>::createNew ( vDim * hDim, vDim * hDim );
  for ( int a = 0; a < vDim; ++a) {
    for ( int b = 0; b < hDim; ++b) {
      for ( int c = 0; c < vDim; ++c) {
        for ( int d = 0; d < hDim; ++d) {

          // compute the covariances
          double EWab  = 0.;
          double EWcd  = 0.;
          double EWabWcd = 0.;

          std::vector<Sample>::const_iterator it;
          for (it = samples.begin(); it != samples.end(); ++it) {
            double EWabUpdate = it->v (a) * it->h (b);
            double EWcdUpdate = it->v (c) * it->h (d);
            EWab  += EWabUpdate;
            EWcd  += EWcdUpdate;
            EWabWcd += EWabUpdate * EWcdUpdate;
          }

          // normalize expectations
          EWab /=  samples.size();
          EWcd /=  samples.size();
          EWabWcd /= samples.size();

          // set Fisher ( Wab, Wcd )
          int i = a * hDim + b;
          int j = c * hDim + d;
          FWW (i, j) = EWabWcd - EWab * EWcd;
        }
      }
    }
  }

  // FWVB = COV ( Wab, visibleBias(j) )
  clap::Matrix<double> FWVB = clap::Matrix<double>::createNew ( vDim * hDim, vDim );
  for ( int a = 0; a < vDim; ++a) {
    for ( int b = 0; b < hDim; ++b) {
      for ( int j = 0; j < vDim; ++j) {
        // compute the covariances
        double EWab  = 0.;
        double EVBj  = 0.;
        double EWabVBj = 0.;

        std::vector<Sample>::const_iterator it;
        for (it = samples.begin(); it != samples.end(); ++it) {
          double EWabUpdate = it->v (a) * it->h (b);
          double EVBjUpdate = it->v (j);
          EWab  += EWabUpdate;
          EVBj  += EVBjUpdate;
          EWabVBj += EWabUpdate * EVBjUpdate;
        }

        // normalize expectations
        EWab /=  samples.size();
        EVBj /=  samples.size();
        EWabVBj /= samples.size();

        // set Fisher ( Wab, Wcd )
        int i = a * hDim + b;
        FWVB (i, j) = EWabVBj - EWab * EVBj;
      }
    }
  }

  // FWHB = COV ( Wab, hiddenBias(j) )
  clap::Matrix<double> FWHB = clap::Matrix<double>::createNew ( vDim * hDim, hDim );
  for ( int a = 0; a < vDim; ++a) {
    for ( int b = 0; b < hDim; ++b) {
      for ( int j = 0; j < hDim; ++j) {
        // compute the covariances
        double EWab  = 0.;
        double EHBj  = 0.;
        double EWabHBj = 0.;

        std::vector<Sample>::const_iterator it;
        for (it = samples.begin(); it != samples.end(); ++it) {
          double EWabUpdate = it->v (a) * it->h (b);
          double EHBjUpdate = it->h (j);
          EWab  += EWabUpdate;
          EHBj  += EHBjUpdate;
          EWabHBj += EWabUpdate * EHBjUpdate;
        }

        // normalize expectations
        EWab /=  samples.size();
        EHBj /=  samples.size();
        EWabHBj /= samples.size();

        // set Fisher ( Wab, Wcd )
        int i = a * hDim + b;
        FWHB (i, j) = EWabHBj - EWab * EHBj;
      }
    }
  }

  // FVBVB = COV ( visibleBias(i), visibleBias(j) )
  clap::Matrix<double> FVBVB = clap::Matrix<double>::createNew ( vDim, vDim );
  for ( int i = 0; i < vDim; ++i) {
    for ( int j = 0; j < vDim; ++j) {
      // compute the covariances
      double EVBi  = 0.;
      double EVBj  = 0.;
      double EVBiVBj = 0.;

      std::vector<Sample>::const_iterator it;
      for (it = samples.begin(); it != samples.end(); ++it) {
        double EVBiUpdate = it->v (i);
        double EVBjUpdate = it->v (j);
        EVBi  += EVBiUpdate;
        EVBj  += EVBjUpdate;
        EVBiVBj += EVBiUpdate * EVBjUpdate;
      }

      // normalize expectations
      EVBi /=  samples.size();
      EVBj /=  samples.size();
      EVBiVBj /= samples.size();

      // set Fisher ( Wab, Wcd )
      FVBVB (i, j) = EVBiVBj - EVBi * EVBj;
    }
  }

  // FVBHB = COV ( visibleBias(i), hiddenBias(j) )
  clap::Matrix<double> FVBHB = clap::Matrix<double>::createNew ( vDim, hDim );
  for ( int i = 0; i < vDim; ++i) {
    for ( int j = 0; j < hDim; ++j) {
      // compute the covariances
      double EVBi  = 0.;
      double EHBj  = 0.;
      double EVBiHBj = 0.;

      std::vector<Sample>::const_iterator it;
      for (it = samples.begin(); it != samples.end(); ++it) {
        double EVBiUpdate = it->v (i);
        double EHBjUpdate = it->h (j);
        EVBi  += EVBiUpdate;
        EHBj  += EHBjUpdate;
        EVBiHBj += EVBiUpdate * EHBjUpdate;
      }

      // normalize expectations
      EVBi /=  samples.size();
      EHBj /=  samples.size();
      EVBiHBj /= samples.size();

      // set Fisher ( Wab, Wcd )
      FVBHB (i, j) = EVBiHBj - EVBi * EHBj;
    }
  }

  // FHBHB = COV ( hiddenBias(i), hiddenBias(j) )
  clap::Matrix<double> FHBHB = clap::Matrix<double>::createNew ( hDim, hDim );
  for ( int i = 0; i < hDim; ++i) {
    for ( int j = 0; j < hDim; ++j) {
      // compute the covariances
      double EHBi  = 0.;
      double EHBj  = 0.;
      double EHBiHBj = 0.;

      std::vector<Sample>::const_iterator it;
      for (it = samples.begin(); it != samples.end(); ++it) {
        double EHBiUpdate = it->h (i);
        double EHBjUpdate = it->h (j);
        EHBi  += EHBiUpdate;
        EHBj  += EHBjUpdate;
        EHBiHBj += EHBiUpdate * EHBjUpdate;
      }

      // normalize expectations
      EHBi /=  samples.size();
      EHBj /=  samples.size();
      EHBiHBj /= samples.size();

      // set Fisher ( Wab, Wcd )
      FHBHB (i, j) = EHBiHBj - EHBi * EHBj;
    }
  }

  // Copy everything to F
  int thetaSize = hDim * vDim + vDim + hDim;
  clap::Matrix<double> F = clap::Matrix<double>::createNew ( thetaSize, thetaSize );
  for ( int i = 0; i < FWW.rows(); ++i) {
    for ( int j = 0; j < FWW.cols(); ++j) {
      F (i, j) = FWW (i, j);
    }
  }
  for ( int i = 0; i < FWVB.rows(); ++i) {
    for ( int j = 0; j < FWVB.cols(); ++j) {
      double val = FWVB (i, j);
      int joffset = FWW.cols();
      F (i, joffset + j) = val;
      F (joffset + j, i) = val;
    }
  }
  for ( int i = 0; i < FWHB.rows(); ++i) {
    for ( int j = 0; j < FWHB.cols(); ++j) {
      double val = FWHB (i, j);
      int joffset = FWW.cols() + FWVB.cols();
      F (i, joffset + j) = val;
      F (joffset + j, i) = val;
    }
  }
  for ( int i = 0; i < FVBVB.rows(); ++i) {
    for ( int j = 0; j < FVBVB.cols(); ++j) {
      double val = FVBVB (i, j);
      int ioffset = FWW.rows();
      int joffset = FWW.cols();
      F (ioffset + i, joffset + j) = val;
      F (joffset + j, ioffset + i) = val;
    }
  }
  for ( int i = 0; i < FVBHB.rows(); ++i) {
    for ( int j = 0; j < FVBHB.cols(); ++j) {
      double val = FVBHB (i, j);
      int ioffset = FWW.rows();
      int joffset = FWW.cols() + FVBVB.cols();
      F (ioffset + i, joffset + j) = val;
      F (joffset + j, ioffset + i) = val;
    }
  }
  for ( int i = 0; i < FHBHB.rows(); ++i) {
    for ( int j = 0; j < FHBHB.cols(); ++j) {
      double val = FHBHB (i, j);
      int ioffset = FWW.rows() + FVBVB.rows();
      int joffset = FWW.cols() + FVBVB.cols();
      F (ioffset + i, joffset + j) = val;
      F (joffset + j, ioffset + i) = val;
    }
  }

  return F;
}

TEST(NaturalGradient,fillSamples) {
  int vDim = 50, hDim = 10;
  unsigned int sampleSize = 500;
  NaturalGradient ng(vDim,hDim,10,sampleSize,sampleSize,1,true);
  clap::RBM<double> rbm (vDim, hDim);
  std::vector<NaturalGradient::Sample> samples;
  ng.fillSamples (rbm,rbm,sampleSize, &samples);
  EXPECT_EQ(sampleSize, samples.size());
}

TEST(NaturalGradient,weightSamples) {
  int vDim = 50, hDim = 1;
  NaturalGradient ng(vDim,hDim,10,4,4,1,true);
  std::vector<NaturalGradient::Sample> samples;
  clap::Matrix<double> v = clap::Matrix<double>::createZero(vDim,1);
  clap::Matrix<double> h = clap::Matrix<double>::createZero(hDim,1);
  samples.push_back( NaturalGradient::Sample(v, h, 100) ); // W = 0
  samples.push_back( NaturalGradient::Sample(v, h, 50) );  // W = 1/2
  samples.push_back( NaturalGradient::Sample(v, h, 50) );  // W = 1/2
  samples.push_back( NaturalGradient::Sample(v, h, 0) );   // W = 1
  BinaryRankingFunctor binaryRanking(samples.size() / 2);
  std::vector<double> W = ng.weightSamples (binaryRanking,samples);
  EXPECT_EQ(0, W[0]);
  EXPECT_EQ(.5, W[1]);
  EXPECT_EQ(.5, W[2]);
  EXPECT_EQ(1,W[3]);
}

TEST(NaturalGradient,computeFisher) {
  int vDim = 10, hDim = 10;
  NaturalGradient ng(vDim,hDim,50,500,500,1,true);
  clap::RBM<double> rbm (vDim, hDim);
  std::vector<NaturalGradient::Sample> samples;
  ng.fillSamples (rbm, rbm, 500, &samples);
  clap::Matrix<double> Finv1, Finv2;
  clap::Matrix<double> Ex2Test, ExTest;
  Finv1 = ng.computeFisher (samples,Ex2Test, ExTest);
  Finv2 = ng.computeFisher2 (samples);
  EXPECT_TRUE( Finv1.isNear(Finv2,0.000001) );
}



